#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Provide a library with fast transformer implementations."""

__author__ = "Angelos Katharopoulos, Apoorv Vyas"
__copyright__ = "Copyright (c) 2020 Idiap Research Institute"
__license__ = "MIT"
__maintainer__ = "Angelos Katharopoulos, Apoorv Vyas"
__email__ = "angelos.katharopoulos@idiap.ch, avyas@idiap.ch"
__url__ = "https://github.com/idiap/fast-transformers"
__version__ = "0.2.1"
